#use "list.apl".
#open List.

id : name_type.
tm : type.
ty : type.

var : id -> tm.
app : tm -> tm -> tm.
lam : id\tm -> tm.
==> : ty -> ty -> ty.
infixr ==> 5.


pred fvs tm [id].
fvs (var V) [V].
fvs (app E1 E2) L3 :- fvs E1 L1, fvs E2 L2, append (L1,L2,L3).
fvs (lam (x\E)) L2 :- fvs E L1, remove(x,L1,L2).

pred freshen tm tm.
freshen (var X) (var X).
freshen (app E1 E2) (app E1' E2') :- freshen E1 E2, freshen E1' E2'.
freshen (lam (x\E)) (lam (y\E'')) / y # E :- freshen E E', E'' = (x~y) E'.

pred eq tm tm.
eq (var X) (var X).
eq (app E1 E2) (app E1' E2') :- eq E1 E2, eq E1' E2'.
eq (lam (x\E)) (lam (x\E')) :- eq E E'.

(* correct substitution *)


pred subst (id\tm) tm tm.
subst (a\var a) E E.
subst (a\var B) E (var B) :- a # B.
subst (a\app E1 E2) E (app E1' E2') :- subst (a\E1) E E1', subst (a\E2) E E2'.
subst (a\lam (b\E1)) E (lam (b\E1')) / b # E :- subst (a\E1) E E1'.

(* declarative substitution (needs -hh -s) *)
func subst1 id tm tm = tm.
subst1 a (var a) E = E.
subst1 a (var b) E = (var b).
subst1 a (app E1 E2) E = (app (subst1 a E1 E) (subst1 a E2 E)).
subst1 a (lam (b\E1)) E = (lam(b\subst1 a E1 E)) :- b # E .


pred beta tm tm.
beta (app (lam (b\E)) E') E'' :- subst(b\E) E' E''.

pred beta' tm tm.
beta' (app (lam (b\E)) E') E'' :- subst1 b E E' = E''.

?beta (app (lam (a\ lam (b\ var a))) (var b)) (lam (c\var b)).
?beta (app (lam (a\ lam (b\ var a))) (var b)) (lam (b\var b)).
?beta (app (lam (a\ lam (b\ var a))) (var b)) (lam (b\var a)).
?beta' (app (lam (a\ lam (b\ var a))) (var b)) (lam (c\var b)).
?beta' (app (lam (a\ lam (b\ var a))) (var b)) (lam (b\var b)).
?beta' (app (lam (a\ lam (b\ var a))) (var b)) (lam (b\var a)).

pred eta tm tm.
eta (lam (x\app E (var x))) E / x # E.


type ctx = [(id,ty)].

pred typeof ctx tm ty.
typeof Gamma (var V) T :- mem((V,T),Gamma).
typeof Gamma (app E1 E2) T 
	:- typeof Gamma E1 (T0 ==> T), 
           typeof Gamma E2 T0.

typeof Gamma (lam(x\E)) (T1 ==> T2) / x # Gamma
	:- typeof ((x,T1)::Gamma) E T2.

(* single step reduction *)
pred step tm tm.
step M M' :- beta M M'.
step (app M N) (app M' N) :- step M M'.
step (app M N) (app M N') :- step N N'.
step (lam (x\M)) (lam (x\M')) :- step M M'.

pred steps tm tm.
steps M M' :- beta M M'.
steps (var X) (var X).
steps (app M N) (app M' N') :- steps M M', steps N N'.
steps (lam (x\M)) (lam (x\M')) :- steps M M'.

?lam(y\Y) = lam(x\Y).

?lam(a\ lam(b\ app X1 (var b))) = lam(b\ lam(a\ app (var a) X1)).
?lam(a\ lam(b\ app X2 (var b))) = lam(b\ lam(a\ app (var a) X3)).
?lam(a\ lam(b\ app (var b) X4)) = lam(b\ lam(a\ app (var a) X5)).
?lam(a\ lam(b\ app (var b) X6)) = lam(a\ lam(a\ app (var a) X7)).

?typeof [] (lam(x\lam(x\var x))) T.

?typeof [] (lam (f\(lam (x\ app(var x) (lam (y\ var y)))))) T.

?typeof [] (lam (x\ app(var x) (var x)))T.

?typeof [(x,T)] (app (var x) (var x)) T.

