open Types;;
open Perm;;
open Printer;;
open Fixity;;

type name_sort = Nstbl.path;;
type data_sort = Nstbl.path;;

type kind = TypeK
  | NameK (* illegal as result kind *)
  | ArrowK of kind * kind
;;


type ty = IdTy of Nstbl.path
  | NameTy of name_sort
  | DataTy of data_sort * ty list
  | UnitTy
  | PairTy of ty * ty
  | AbsTy of ty * ty
  | ListTy of ty
  | VarTy of var
  | UnderscoreTy 
  | IntTy
  | CharTy
  | BoolTy
  | PropTy
  | ArrowTy of ty * ty
;;


type constr = (name*var) list;;


type atomic = Nstbl.path * term list
and term = Atomic of atomic
  | Var of var
  | Name of name
  | Transpose of name * name * term
  | Nil
  | Cons of term * term
  | Unit
  | Pair of term * term
  | Abs of name * term
  | True
  | Fresh of term * term
  | Eq of term * term
  | Is of term * term
  | Implies of term * term
  | Or of term * term
  | And of term * term
  | Forall of var * ty * term
  | Exists of var * ty * term
  | New of var * ty * term 
  | Cut 
  | Guard of term * term * term
  | Not of term
  | Constr of term * constr
  | Underscore
  | IntC of int
  | CharC of char
  | BoolC of bool
;;


type decl = {pos:Pos.pos option; rdecl:rdecl}
and rdecl = KindDecl of sym * kind
  | SymDecl of sym * ty
  | FuncDecl of sym * ty list * ty
  | PredDecl of sym * ty list
  | ClauseDecl of term 
  | Query of term
  | UseDirective of string
  | TraceDirective of int
  | InfixDecl of direction * string * int
  | NamespaceDecl of sym * decl list
  | TypeDefn of sym * var list * ty 
  | BuiltinFuncDecl of sym * ty list * ty 
	* (Isym.term_sym Internal.term list -> Isym.term_sym Internal.term)
  | TypeQuery of term
  | QuitDirective
  | OpenDirective of Nstbl.path
  | HelpDirective of Nstbl.path option
;;


val pp_kind : kind printer;;
val pp_term : term printer;;
val pp_ty : ty printer;;
val pp_decl : decl printer;;

val k2s : kind -> string;;
val t2s : term -> string;;
val ty2s : ty -> string;;

val ftvs    : ty -> Varset.t;;

val freshen    : ty -> ty;;

val supp : term -> Varset.t;;

val apply_tysub : ty Varmap.t -> ty -> ty;;
