(* Alpha Prolog *)
(* Internal representation of untyped terms, goals, and program clauses. *)

open Types;;
open Printer;;

type perm = Perm.perm;;


type bconstr = Toplevel
             | Constrained of var list
;;


type bind = Exist of bconstr
          | Univ
;;


type 'a term = App of 'a * 'a term list
  | Abs of  name * 'a term
  | Name of name
  | Susp of perm * bind * var 
;;


type 'a goal = 
(* propositions *)
    Gtrue
  | Gatomic of 'a term
  | Gand of 'a goal * 'a goal
  | Gor of 'a goal * 'a goal
  | Gimplies of 'a prog * 'a goal
(* first-order quantifiers *)
  | Gforall of var * 'a goal
  | Gexists of var * 'a goal
  | Gnew of name * 'a goal
(* term predicates *)
  | Gfresh of 'a term * 'a term
  | Gequals of 'a term * 'a term
  | Gis of 'a term * 'a term 
(* control *)
  | Gcut
  | Guard of 'a goal * 'a goal * 'a goal
  | Gnot of 'a goal

and 'a prog = 
(* first-order program clauses *)
    Dtrue
  | Datomic of 'a term
  | Dimplies of 'a goal * 'a prog
  | Dforall of var * 'a prog
  | Dand of 'a prog *'a  prog
(* "new" quantifier *)
  | Dnew of name * 'a prog
;;


val mk_var : var -> 'a term;;

val pp_term : 'a printer -> 'a term printer;;
val pp_goal : 'a printer -> 'a goal printer;;
val pp_prog : 'a printer -> 'a prog printer;;


val supp : 'a term -> Varset.t;;
val fas  : 'a term -> Varset.t;;
val fas_g : 'a goal -> Varset.t;;
val fas_p : 'a prog -> Varset.t;;
val fvs  : 'a term -> Varset.t;;
val fvs_g : 'a goal -> Varset.t;;
val fvs_p : 'a prog -> Varset.t;;

val apply_p'  : perm -> 'a term -> 'a term;; (* no suspensions *)
val apply_p   : perm -> 'a term -> 'a term;;
val apply_p_g : perm -> 'a goal -> 'a goal;;
val apply_p_p : perm -> 'a prog -> 'a prog;;


val subst_name : name -> name -> 'a term -> 'a term;;
val subst_name_g : name -> name -> 'a goal -> 'a goal;;
val subst_name_p : name -> name -> 'a prog -> 'a prog;;
