(* Alpha-Prolog *)
(* Permutations *)
open Types;;
open Printer;;

type perm = (name*name) list;;


let id = [];;


let trans a b = [(a,b)];;


let comp p q = p @ q;;


let inv p = List.rev p;;


let rec domain p = 
  match p with 
    [] -> []
  | (a,b)::p -> a::b::domain p
;;


let rec lookup p a = 
  match p with
    [] -> a
  | (b,c)::p' -> 
      let a' = lookup p' a in
      if Var.eq a' b then c
      else if Var.eq a' c then b
      else a'
;;


let apply p p' = 
  List.map (fun (a,b) -> (lookup p a, lookup p b)) p'
;;


let disagreement_set p1 p2 = 
  let dom1 = domain p1 in
  let dom2 = domain p2 in
  let dom = dom1 @ dom2 in
  let rng = List.map 
      (fun x -> x, lookup p1 x, lookup p2 x) dom in
  List.map (fun (x,y,z) -> x) 
    (List.filter (fun (x,y,z) -> not(Var.eq y z)) rng)
;;


let rec pp_perm p = 
  match p with 
    [] -> emp
  | (a,b)::p -> text "(" <:> 
      text (Var.to_string' a) <:> 
      text " " <:>
      text (Var.to_string' b) <:>
      text ")"
;;


let from_list l = l;;


let to_list p = p;;


let p2s p = Printer.print_to_string pp_perm p;;
