open Util;;
open Internal;;
open Subst;;
open Types;;
open Unify;;
module P = Perm;;
module S = Subst;;

(* TODO: Faster unification *)

(* TODO: Research correct way to handle new-variables and quantification
   (and discharge freshness constraints)
*)


type 'a answer =  'a Unify.answer ;;


let remove_name_dependences a (subst,constrs) = 
  (subst, List.filter (fun (av,v) -> 
    match av with
      Name b -> if Var.eq a b then false else true
    | _ -> true) constrs)
;;


let remove_var_dependences x (subst,constrs) = 
  (subst, List.filter (fun (av,v) -> 
    if Var.eq x v then false else 
    match av with
      Susp(_,_,w) -> if Var.eq x w then false else true
    | _ -> true) constrs)
;;


let check_for_var_dependences x (subst,constrs) = 
  List.exists (fun (av,v) -> 
    if Var.eq x v then true 
    else match av with
      Susp(_,_,w) -> Var.eq x w
    | _ -> false) constrs
;;


let solve_wrap pp idx prog goal sc ans = 
  let rec solve_goal prog goal uvars names cexn sc ans = 
    do_trace 2 (fun () -> 
      if goal != Gtrue then 
	let (subst,constrs) = ans in
	print_string ("Solving goal\n"^
		      Printer.print_to_string (pp_goal pp) 
			(Subst.apply_s_g subst goal)^
		      "\n"));
    let b = (match goal with
      Gtrue -> sc ans
    | Gand(g1,g2) ->  
	solve_goal prog g1 uvars names cexn 
	  (solve_goal prog g2 uvars names cexn sc) 
	  ans
    | Gor(g1,g2) -> 
	solve_goal prog g1 uvars names cexn sc ans || 
	solve_goal prog g2 uvars names cexn sc ans
    | Gatomic(t) -> 
	(let module M = struct exception NewCut end in 
	try (match_atomic t prog uvars names M.NewCut sc ans)
	with M.NewCut -> true) (* could this be false? *)
    | Gexists(x,g) -> 
	let y = Var.rename x in
	let t = Susp(Perm.id,Exist(Constrained (names@uvars)),y) in
	solve_goal prog (S.apply_s_g (S.sub1 x t) g) uvars names cexn 
	  (fun ans -> sc (remove_var_dependences y ans)) ans
    | Gforall(x,g) -> 
	let x' = Var.rename x in
	solve_goal 
	  prog (S.apply_s_g (S.sub1 x (Susp(Perm.id,Univ,x'))) g) 
          (x'::uvars) names cexn 
	  (fun ans -> if check_for_var_dependences x' ans then true else sc ans)
	  ans
	(* TODO: limit names available to evars also! *)
    | Gnew(a,g) ->  
	let b = Var.rename a in
	if !Flags.substitute_names 
	then 
	  solve_goal prog (subst_name_g a b g) uvars (b::names) 
	    cexn (fun ans -> sc (remove_name_dependences b ans)) ans
	else 
	  solve_goal prog (apply_p_g (P.trans a b) g) uvars (b::names) 
	    cexn (fun ans -> sc (remove_name_dependences b ans)) ans
    | Gimplies(d,g) -> 
	solve_goal (Dand(d,prog)) g uvars names cexn sc ans
    | Gequals(t1,t2) -> 
	let (ans') = ans in
	unify Isym.term_sym_eq t1 t2 (fun ans' -> sc (ans')) ans'
    | Gis(t1,t2) -> 
	let ((subst,constrs)) = ans in
	let v = Eval.eval Runtime.apply subst t2 in 
	unify Isym.term_sym_eq t1 v (fun ans' -> sc (ans')) (subst,constrs)
    | Gfresh (t1,t2) -> 
	let (ans') = ans in
	unify_fresh t1 t2 (fun ans' -> sc (ans')) ans'
    | Gnot g -> 
	let module M = struct exception Not_success end in
	let new_sc = (fun ans -> raise M.Not_success) in
	(try 
	  let _ = solve_goal prog g uvars names cexn new_sc ans in
	  sc ans
	with M.Not_success -> false) 
    | Guard(g1,g2,g3) -> 
	let module M = 
	  struct exception Guard_success of Isym.term_sym answer end in
	let new_sc = (fun ans -> raise (M.Guard_success ans)) in
	let _ = (try (
	  let _ = solve_goal prog g1 uvars names cexn new_sc ans in
	  solve_goal prog g3 uvars names cexn sc ans )
	with M.Guard_success ans ->
	  solve_goal prog g2 uvars names cexn sc ans)
	in true (* could this be false? *)
    | Gcut -> 
	let _ = sc ans in
	raise cexn
    ) in
    do_trace 2 (fun () -> 
      print_string "Backtracking solve_goal\n");
    b
      
  and match_atomic t prog uvars names cexn sc ans = 
    let prog' = Index.lookup idx t in
    let rec rmatch p sg names = 
      match p with 
	Dtrue -> false
      | Dand(d1,d2) -> 
	  rmatch d1 sg names; 
	  rmatch d2 sg names;
	  true
      | Dimplies(g,s) -> 
	  rmatch s (Gand(sg,g)) names
      | Datomic(s) -> 
	  do_trace 3 (fun () -> 
	    let (subst,constrs) = ans in 
	    print_string "Unifying with clause head\n";
	    Printer.print_to_channel 
	      (pp_term pp) 
	      (Subst.apply_s subst s) stdout;
	    print_string "\n");
	  let b = unify Isym.term_sym_eq s t 
	      (solve_goal prog sg uvars names cexn sc) 
	      ans in
    	  do_trace 3 (fun () -> 
	    print_string "Unification failed.\n");
	  b
      | Dforall(x,d) -> 
	  rmatch 
	    (S.apply_s_p 
	       (S.sub1 x 
		  (Susp(Perm.id,Exist(Constrained (names@uvars)),Var.rename x))) d) 
	    sg names
      | Dnew(a,d) -> 
	  (* TODO: Deal with freshness constraints *)
	  let b = Var.rename a in
	  do_trace 2 (fun () -> 
	    print_string ("Trying fresh name "^Var.to_string' b^" in new\n")
	      );
	  let do_inst b = 
	    if !Flags.substitute_names
	    then subst_name_p a b d
	    else apply_p_p (P.trans a b) d
	  in 
	  let name_mismatch = rmatch (do_inst b) sg (b::names) in
	  (* Try existing fresh names if gensym'd one fails. *)
	  (* TODO: We may be being way too paranoid about freshness checks,
	     name-name checks needed when names are instantiated using 
	     substitution but not when using permutation. *)
	  if (!Flags.search_names) && (name_mismatch || not(!Flags.fast_search) )
	  then 
	    (do_trace 2 (fun () -> 
	      print_string( "Trying ["
			   ^string_of_int (List.length names)
			   ^"] existing names in new with atom_mismatch = "^
			   (if name_mismatch then "true" else "false")^"\n")
		);
	     let mk_g c = 
	       if !Flags.strict_new_freshness 
	       then 
		 let fvs_p = fvs_p p in
		 let fas_p = fas_p p in (* use p, not d, to exclude a. *)
	       	 let fgs = List.map (fun v -> 
		   Gfresh(Name c,Susp(Perm.id,Univ,v)))
		     (Varset.elements fvs_p) in
		 let fg = List.fold_left (fun g g' -> Gand(g,g')) Gtrue fgs in
		 let fgs' = List.map (fun a -> 
		   Gfresh(Name c,Name(a)))
		     (Varset.elements fas_p) in
		 let fg' = List.fold_left 
		     (fun g g' -> Gand(g,g')) Gtrue fgs' in
		 Gand(Gand(sg,fg),fg')
	       else sg
	     in 
	     List.iter (fun c -> 
	       let _ = rmatch (do_inst c) (mk_g c) names in 
	       ())
	       names;
	     true)
	  else name_mismatch
    in
    do_trace 1 (fun () -> 
      let (subst,constrs) = ans in
      print_string "Solving atomic goal\n";
      Printer.print_to_channel (pp_term pp) (Subst.apply_s subst t) stdout;
      print_string "\n");
    let b = rmatch (Dand (prog',prog)) Gtrue names in
    do_trace 1 (fun () -> 
      print_string "Backtracking match_atomic\n");
    b
  in 
  let module M = struct exception Cut end in
  (try (let names = Varset.elements (fas_g goal) in 
        solve_goal prog goal [] names M.Cut sc ans)
  with M.Cut -> false)
;;
  

let solve pp idx goal sc = 
  Var.reset_var();
  solve_wrap pp idx Dtrue goal sc (S.id,[])
;;
